#include <PID_v1.h>

byte encoder_Port_D;      //I'm using direct port manipulation. I use these variables for storing the state of Port D.
byte last_encoder_Port_D;
byte stepDir_Port_D = 0;

volatile int encoderbuffer = 0;
volatile int layeredencoderbuffer = 0;
volatile bool encodererror = 0;

volatile int interruptbuffer = 0;
volatile int layeredbuffer = 0;

long ctrldiff = 0;

double pidSetpoint = 0; //In order to never overflow the Setpoint stays 0 and the step input is handeled as deviation.
double pidInput;
double pidOutput;

//PID parameters are determinated experimentaly
double Kp=20, Ki=0.01, Kd=0.07;
PID myPID(&pidInput, &pidOutput, &pidSetpoint, Kp, Ki, Kd, DIRECT);

void setup() {

  pinMode(8, OUTPUT); //Dir output
  pinMode(9, OUTPUT); //Run output
  
  //setup timer interrupt
  cli();//stop interrupts
  
  //set timer2 interrupt at 8kHz
  TCCR2A = 0;// set entire TCCR2A register to 0
  TCCR2B = 0;// same for TCCR2B
  TCNT2  = 0;//initialize counter value to 0
  // set compare match register for 8khz increments
  OCR2A = 249;// = (16*10^6) / (8000*8) - 1 (must be <256)
  // turn on CTC mode
  TCCR2A |= (1 << WGM21);
  // Set CS21 bit for 8 prescaler
  TCCR2B |= (1 << CS21);   
  // enable timer compare interrupt
  TIMSK2 |= (1 << OCIE2A);

  sei();//allow interrupts

  //STEP input pin Interrupt
  attachInterrupt(digitalPinToInterrupt(3), isr_in, RISING);

  //settings for the PID library
  myPID.SetSampleTime(1);
  myPID.SetOutputLimits(-255, 255);
  myPID.SetMode(AUTOMATIC);
}

void loop() {

  cli();//stop interrupts to read the buffers for the encoder and Step/Dir input
  layeredencoderbuffer = encoderbuffer;
  encoderbuffer -= layeredencoderbuffer;

  layeredbuffer = interruptbuffer;
  interruptbuffer -= layeredbuffer;
  sei();//allow interrupts

  //add the two buffers contents to the deviation
  ctrldiff += layeredencoderbuffer;
  ctrldiff += layeredbuffer;
  
  pidInput = ctrldiff;
  
  myPID.Compute();
  
  setMotor(pidOutput);
  delay(1);
}

//takes a value from -255 to 255 and runs the motor accordingly
void setMotor(int mValue){
  mValue = constrain(mValue,-255,255);

  if(mValue == 0){
    digitalWrite(9, LOW);
  }
  else if(mValue > 0){
    analogWrite(9, mValue);
    digitalWrite(8, LOW); //direction
  }
  else if(mValue < 0){
    analogWrite(9, -mValue);
    digitalWrite(8, HIGH);  //direction
  }
  return;
}

//Interrupt on pin 3 for the Step/Dir input
void isr_in() {
  stepDir_Port_D = PIND;  //read and mask the whole port for speed
  stepDir_Port_D &= B00010000;
  if(stepDir_Port_D == B00010000){
    interruptbuffer -= 1;
  }
  else{
    interruptbuffer += 1;
  }
}

//Timer interrupt at 8kHz for reading the hall sensors as encoder
ISR(TIMER2_COMPA_vect){
  encoder_Port_D = PIND;  //read and mask the whole port for speed
  encoder_Port_D &= B11100000;
  if(encoder_Port_D != last_encoder_Port_D){  //save time by only acting if anything changed
    switch (last_encoder_Port_D) {  //layered switch case to read the sensors there might be a better way to do this but it works
      case B10000000:
      switch (encoder_Port_D) {
        case B11000000:
        encoderbuffer --;
        break;
        case B10100000:
        encoderbuffer ++;
        break;
        default:
        encodererror = 1;
        break;
      }
      break;
      case B11000000:
      switch (encoder_Port_D) {
        case B01000000:
        encoderbuffer --;
        break;
        case B10000000:
        encoderbuffer ++;
        break;
        default:
        encodererror = 1;
        break;
      }
      break;
      case B01000000:
      switch (encoder_Port_D) {
        case B01100000:
        encoderbuffer --;
        break;
        case B11000000:
        encoderbuffer ++;
        break;
        default:
        encodererror = 1;
        break;
      }
      break;
      case B01100000:
      switch (encoder_Port_D) {
        case B00100000:
        encoderbuffer --;
        break;
        case B01000000:
        encoderbuffer ++;
        break;
        default:
        encodererror = 1;
        break;
      }
      break;
      case B00100000:
      switch (encoder_Port_D) {
        case B10100000:
        encoderbuffer --;
        break;
        case B01100000:
        encoderbuffer ++;
        break;
        default:
        encodererror = 1;
        break;
      }
      break;
      case B10100000:
      switch (encoder_Port_D) {
        case B10000000:
        encoderbuffer --;
        break;
        case B00100000:
        encoderbuffer ++;
        break;
        default:
        encodererror = 1;
        break;
      }
      break;
      default:
      break;
    }
    last_encoder_Port_D = encoder_Port_D;
  }
}
